use strict;
use Test::More 0.98;

use_ok $_ for qw(
    Getopt::EX
    Getopt::EX::Colormap
    Getopt::EX::Func
    Getopt::EX::LabeledParam
    Getopt::EX::Loader
    Getopt::EX::Long
    Getopt::EX::Module
);

done_testing;

